# Load libraries
import pandas as pd
import scanpy as sc
import numpy as np
import scanpy.external as sce



class MakeAnndata:
    # This class is used to make the anndata with spatial coordinates from CURIO.
    def __init__(self, sample_path, barcodes_path):
        # sample_path = the path to the h5ad file generated by CURIO pipeline.
        # barcodes_path = the path to the bead barcode whitelist corresponding to sample.
        self.sample = sc.read_h5ad(sample_path)
        self.barcodes = pd.read_csv(barcodes_path, delimiter="\t", names=["barcode", "x", "y"])
        # the data will be the finished anndata.
        self.data = self._shared_barcodes()

    def _process_whitelist(self):
        # Function to preprocess bead barcode whitelist.
        print(self.barcodes[["x", "y"]].to_numpy())
        barcodes = self.barcodes
        barcodes.index = barcodes["barcode"]
        return barcodes

    def _process_h5ad(self):
        # Function to preprocess h5ad file from CURIO.
        sc.pp.filter_cells(self.sample, min_counts=1)
        data = self.sample
        data.obs_names = [(''.join(x.split("_"))) for x in data.obs_names]
        return data

    def _shared_barcodes(self):
        # Function to generate finale anndata file with spatial positions.
        barcodes = self._process_whitelist()
        data = self._process_h5ad()

        shared_barcodes = list(set(data.obs_names).intersection(set(barcodes["barcode"])))
        data = data[shared_barcodes]
        barcodes = barcodes.reindex(data.obs_names)
        data.obs = data.obs.join(barcodes)
        data.obsm["spatial"] = barcodes[["x", "y"]].to_numpy()
        return data


import matplotlib.pyplot as plt

from .MakeAnndata import MakeAnndata
from .remove_mt_ribo import remove_mt_ribo
from matplotlib_venn import venn3
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import scanpy as sc
import scanpy.external as sce
class analysation:
    def __init__(self,h5ad_path, barcodes_path, cell_cycle_gene_path, resolution):
        # h5ad_path = path to h5ad file generated by CURIO pipeline
        # barcodes_path = path to bead barcode whitelist for specific slide
        self.barcodes_path = barcodes_path
        self.h5ad_path = h5ad_path
        # cell cycle genes
        self.cell_cycle_gene_path = cell_cycle_gene_path

        self.s_genes,self.g2m_genes = self.cell_cycle()

        self.resolution = resolution

        # The mitocondrial statisitcs:
        self.mito_obs, self.mito_data = self.mito_stat()

        self.normalised_data = self.preprocessing_anndata()

        self.final_anndata, self.ranked_genes = self.clustering()

    def mito_stats(self):
        # Function which gives statistics of mitocondrial and ribosomal genes

        tmp = MakeAnndata(self.h5ad_path, self.barcodes_path)

        data = tmp.data

        tmp = remove_mt_ribo(data)

        mito_obs = tmp.mito_obs
        mito_data = tmp.mito_data

        sc.settings.set_figure_params(dpi_save=400, fontsize=30, figsize=(10, 10), facecolor='white', frameon=False,
                                      transparent=True, format="pdf")

        sc.pl.violin(mito_data, ['n_genes_by_counts', 'total_counts', 'pct_counts_mt', 'pct_counts_ribo', 'pct_counts_hb'],
                     jitter=0.4, rotation=45, title= "Violin Plots of Quality Indicators")
        sc.pl.violin(mito_data, ['pct_counts_mt'],
                     jitter=0.4, rotation=45, title = "Violin Plot of Mito Data")
        sc.pl.scatter(mito_data, x='total_counts', y='pct_counts_mt', title= "Scatter Plot of pct_counts_mt")
        sc.pl.scatter(mito_data, x='total_counts', y='n_genes_by_counts', title="Scatter Plot of n_genes_by_counts")

        sc.settings.set_figure_params(dpi_save=400, fontsize=6, figsize=(10, 10), facecolor='white', frameon=False,
                                      transparent=True, format="pdf")


        sc.pl.spatial(
        mito_data,
        color=["total_counts", "n_genes_by_counts", "pct_counts_mt"],
        spot_size=30, vmin=0, vmax='p99.2', save="heart_spatial_before_filtering")

        sc.pl.spatial(
        mito_data,
        color=["pct_counts_mt"],
        spot_size=30, vmin=0, vmax='p99.2', save="heart_spatial_before_filtering")

        data.obs.pct_counts_mt.hist(bins=100)
        plt.title("Histogram of Percentage Counts Mito")

        data.obs.pct_counts_ribo.hist(bins=100)
        plt.title("Histogram of Percentage Counts Ribo")

        return mito_obs, mito_data

    def cell_cycle(self):
        cell_cycle_genes = [x.strip() for x in open(self.cell_cycle_gene_path)]
        # Split into 2 lists
        s_genes = cell_cycle_genes[:43]
        g2m_genes = cell_cycle_genes[43:]

        cell_cycle_genes = [x for x in cell_cycle_genes if x in data.var_names]
        print(len(cell_cycle_genes))

        return s_genes, g2m_genes

    def preprocessing_anndata(self):

        tmp = MakeAnndata(self.h5ad_path, self.barcodes_path)
        data = tmp.data
        tmp = remove_mt_ribo(data)

        data = tmp.data

        sc.pl.highest_expr_genes(data, n_top=20)

        # Normalise - cpm = target_sum = 1e6
        sc.pp.normalize_total(data, inplace=True, target_sum=1e6)

        # Log - OBS! Bug in write_h5ad - see lower down.
        sc.pp.log1p(data)

        # Note that you do want to exclude junk - high expression probably common genes expressed in all cells
        sc.pp.highly_variable_genes(data, max_mean=10, min_mean=0, min_disp=0.25)
        sc.pl.highly_variable_genes(data)

        sc.tl.score_genes_cell_cycle(data, s_genes=self.s_genes, g2m_genes=self.g2m_genes)

        # Select only high var genes
        data.raw = data
        data = data[:, data.var.highly_variable]

        sc.pl.violin(data, ['S_score', 'G2M_score'],
                     jitter=0.4, rotation=45)

        # Scale before pca with exclusion of given sd
        sc.pp.scale(data, max_value=10, zero_center=False)

        # regress out unnecessary sources of variation
        # sc.pp.regress_out(data, ['total_counts', 'pct_counts_mt'])
        # data.raw = data
        sc.pp.pca(data, random_state=0)

        # captured variance
        sc.pl.pca_variance_ratio(data, n_pcs=50, log=True)
        # Calculate neighbour hood graph using pc
        sc.pp.neighbors(data, n_pcs=5)
        # Embedd neighbourhood graph through umap
        sc.tl.umap(data)
        # confirm scaling

        data.to_df().mean(axis=0).hist(bins=100)

        data.to_df().iloc[:, 1000:2000].var(axis=0).hist()

        return data

    def clustering(self):
        data.uns['log1p'] = {'base': None}  # Bug in scanpy! OBS!
        # Leiden clustering
        sc.tl.leiden(data, key_added="leiden_0.3", resolution = 0.3)
        sc.tl.leiden(data, key_added="leiden_0.5", resolution = 0.5)
        sc.tl.leiden(data, key_added="leiden_0.8", resolution = 0.8)
        sc.tl.leiden(data, key_added="leiden_1.0", resolution = 1.0)

        sc.pl.spatial(data, color=["leiden_0.3","leiden_0.5","leiden_0.8", "leiden_1.0"], spot_size=30, wspace=0.5)

        if self.resolution == None:
            sc.pl.spatial(data, color=["leiden_1.0"], spot_size=30, wspace=0.5)
            sc.tl.dendrogram(data, groupby="leiden_1.0")
            sc.pl.dendrogram(data, groupby="leiden_1.0")
            sc.pl.umap(data, color=["leiden_1.0"], palette="tab20")

            sc.tl.rank_genes_groups(data, 'leiden_1.0', method='wilcoxon', key_added='wilcoxon')
            sc.pl.rank_genes_groups(data, n_genes=25, sharey=False, key="wilcoxon")

            sc.tl.rank_genes_groups(data, 'leiden_1.0', method='t-test', key_added='t-test')
            sc.pl.rank_genes_groups(data, n_genes=25, sharey=False, key="t-test")

            sc.tl.rank_genes_groups(data, 'leiden_1.0', method='logreg', key_added="logreg")
            sc.pl.rank_genes_groups(data, n_genes=25, sharey=False, key="logreg")

            sc.tl.rank_genes_groups(data, 'leiden_1.0', method='t-test_overestim_var', key_added="t-test_ov")
            sc.pl.rank_genes_groups(data, n_genes=25, sharey=False, key="t-test_ov")

            ranked_genes = pd.DataFrame(data.uns['wilcoxon']['names'])

            sc.pl.rank_genes_groups_heatmap(data, n_genes=5, key="wilcoxon", groupby="leiden_1.0", show_gene_labels=True)

            sc.pl.rank_genes_groups_dotplot(data, n_genes=5, key="wilcoxon", groupby="leiden_1.0")

            sc.pl.rank_genes_groups_stacked_violin(data, n_genes=5, key="wilcoxon", groupby="leiden_1.0")

            sc.pl.rank_genes_groups_matrixplot(data, n_genes=5, key="wilcoxon", groupby="leiden_1.0")


        else:
            sc.pl.spatial(data,color=[f"leiden_{self.resolution}"], spot_size=30, wspace=0.5)
            sc.tl.dendrogram(data, groupby=f"leiden_{self.resolution}")
            sc.pl.dendrogram(data, groupby=f"leiden_{self.resolution}")
            sc.pl.umap(data, color=[f"leiden_{self.resolution}"], palette="tab20")

            sc.tl.rank_genes_groups(data, f"leiden_{self.resolution}", method='wilcoxon', key_added='wilcoxon')
            sc.pl.rank_genes_groups(data, n_genes=25, sharey=False, key="wilcoxon")

            sc.tl.rank_genes_groups(data, f"leiden_{self.resolution}", method='t-test', key_added='t-test')
            sc.pl.rank_genes_groups(data, n_genes=25, sharey=False, key="t-test")

            sc.tl.rank_genes_groups(data, f"leiden_{self.resolution}", method='logreg', key_added="logreg")
            sc.pl.rank_genes_groups(data, n_genes=25, sharey=False, key="logreg")

            sc.tl.rank_genes_groups(data, f"leiden_{self.resolution}", method='t-test_overestim_var', key_added="t-test_ov")
            sc.pl.rank_genes_groups(data, n_genes=25, sharey=False, key="t-test_ov")

            ranked_genes = pd.DataFrame(data.uns['wilcoxon']['names'])
            sc.pl.rank_genes_groups_heatmap(data, n_genes=5, key="wilcoxon", groupby=f"leiden_{self.resolution}", show_gene_labels = True)

            sc.pl.rank_genes_groups_dotplot(data, n_genes=5, key="wilcoxon", groupby=f"leiden_{self.resolution}")

            sc.pl.rank_genes_groups_stacked_violin(data, n_genes=5, key="wilcoxon", groupby=f"leiden_{self.resolution}")

            sc.pl.rank_genes_groups_matrixplot(data, n_genes=5, key="wilcoxon", groupby=f"leiden_{self.resolution}")

        wc = sc.get.rank_genes_groups_df(data, group='0', key='wilcoxon', pval_cutoff=0.01, log2fc_min=0)['names']
        tt = sc.get.rank_genes_groups_df(data, group='0', key='t-test', pval_cutoff=0.01, log2fc_min=0)['names']
        tt_ov = sc.get.rank_genes_groups_df(data, group='0', key='t-test_ov', pval_cutoff=0.01, log2fc_min=0)['names']


        # compare cluster1 genes, only stores top 100 by default
        venn3([set(wc), set(tt), set(tt_ov)], ('Wilcox', 'T-test', 'T-test_ov'))
        plt.show()

        return data, ranked_genes







from setuptools import setup, find_packages

VERSION = '0.0.5'
DESCRIPTION = 'Python package for analysis of slide-seq data on heart tissue'
LONG_DESCRIPTION = 'A package created for the complete analysis done in a heart tissue slide-seq spatial transcriptomics project.'

# Setting up
setup(
    # the name must match the folder name 'HumanHeartSpatial'
    name="HumanSpatialHeart",
    version=VERSION,
    author="Mercedes Dalman",
    author_email="meraljopedal@gmail.com",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    install_requires=['pandas', 'scanpy', 'seaborn', 'matplotlib', 'numpy'],  # add any additional packages that
    # needs to be installed along with your package. Eg: 'caer'

    keywords=['python', 'spatial transcriptomics', 'slide-seq', 'heart tissue', 'human heart'],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Education",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)


import pandas as pd
import scanpy as sc
import numpy as np
import scanpy.external as sce
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns



class remove_mt_ribo:

    def __init__(self, anndata):
        # anndata = the unprocessed anndata with spatial coordinates generated by MakeAnndata
        self.anndata = anndata

        # metrics for mito genes
        self.mito_obs, self.mito_data = self._before_mt_ribo()

        # Filtered anndata without mt and ribo genes
        self.data = self._remove_mt_ribo()

    def _before_mt_ribo(self):
        # Function to get metrics before filtering
        anndata = self.anndata
        # mitochondrial genes
        anndata.var['mt'] = anndata.var_names.str.startswith('MT-')
        # ribosomal genes
        anndata.var['ribo'] = anndata.var_names.str.startswith(("RPS", "RPL"))
        # hemoglobin genes.
        anndata.var['hb'] = anndata.var_names.str.contains(("^HB[^(P)]"))

        sc.pp.calculate_qc_metrics(anndata, qc_vars=['mt', 'ribo', 'hb'], percent_top=None, log1p=False, inplace=True)

        anndata = self.anndata
        mito_genes = anndata.var_names.str.startswith('MT-')
        # for each cell compute fraction of counts in mito genes vs. all genes
        # the `.A1` is only necessary as X is sparse (to transform to a dense array after summing)
        anndata.obs['percent_mt2'] = np.sum(
            anndata[:, mito_genes].X, axis=1).A1 / np.sum(anndata.X, axis=1).A1
        # add the total counts per cell as observations-annotation to adata
        anndata.obs['n_counts'] = anndata.X.sum(axis=1).A1

        return anndata[:, mito_genes].obs, anndata

    def _remove_mt_ribo(self):
        # Function to filter out noise genes
        anndata = self.anndata
        # remove mt genes, malat1 (non-informative) and hb-genes -
        malat1 = anndata.var_names.str.startswith('MALAT1')
        # need to redefine the mito_genes since they were first calculated on the full object before removing low expressed genes.
        mito_genes = anndata.var_names.str.startswith('MT-')
        hb_genes = anndata.var_names.str.contains('^HB[^(P)]')

        remove = np.add(mito_genes, malat1)
        remove = np.add(remove, hb_genes)
        keep = np.invert(remove)

        anndata = anndata[:, keep]

        return anndata
